import plyj.model as m
import plyj.parser
from abpricou.components import Components

 
class CountingVisitor(m.Visitor):
    def __init__(self):
        super(CountingVisitor, self).__init__()
        self.components = Components()

        self._uninitialised_fields = []
        self._in_ctor = False
        self._uninitialised_locals = []

    def visit_InterfaceDeclaration(self, declaration):
        # ignore interfaces completely, will be implemented somewhere
        return False

    def visit_ClassInitializer(self, initializer):
        self._can_have_final_locals()

        return self._code_is_a_constant()

    def leave_ClassInitializer(self, initializer):
        return self._end_of_method()

    def visit_ConstructorDeclaration(self, declaration):
        self._can_have_final_locals()
        self._in_ctor = True

        self._a_name()
        return self._code_is_a_constant()

    def leave_ConstructorDeclaration(self, declaration):
        self._in_ctor = False
        return self._end_of_method()

    def visit_FieldDeclaration(self, declaration):
        self._find_uninitialised(declaration, self._uninitialised_fields)
        return True

    def visit_VariableDeclaration(self, declaration):
        self._find_uninitialised(declaration, self._uninitialised_locals)
        return True

    def visit_MethodDeclaration(self, declaration):
        self._can_have_final_locals()
        self._a_name()
        return self._code_is_a_constant()

    def leave_MethodDeclaration(self, declaration):
        return self._end_of_method()

    def visit_FormalParameter(self, parameter):
        return self._a_name()

    def visit_VariableDeclarator(self, declarator):
        return self._a_name()

    def visit_Assignment(self, assignment):
        if 'value' in dir(assignment.lhs):
            if self._in_ctor and \
                    assignment.lhs.value in self._uninitialised_fields:
                # a final field is assigned in constructor
                return True
            if assignment.lhs.value in self._uninitialised_locals:
                # a final local is assigned in block
                return True

        if 'name' in dir(assignment.lhs):
            if self._in_ctor and \
                    assignment.lhs.name in self._uninitialised_fields:
                # a final field is assigned in constructor
                return True
            if assignment.lhs.name in self._uninitialised_locals:
                # a final local is assigned in block
                return True

        return self._an_assignment()

    def visit_Conditional(self, conditional):
        return self._a_conditional()

    def visit_ConditionalOr(self, binary):
        return self._a_pseudo_invocation()

    def visit_ConditionalAnd(self, binary):
        return self._a_pseudo_invocation()

    def visit_Or(self, binary):
        return self._a_pseudo_invocation()

    def visit_Xor(self, binary):
        return self._a_pseudo_invocation()

    def visit_And(self, binary):
        return self._a_pseudo_invocation()

    def visit_Equality(self, binary):
        return self._a_pseudo_invocation()

    def visit_InstanceOf(self, binary):
        return self._a_pseudo_invocation()

    def visit_Relational(self, binary):
        return self._a_pseudo_invocation()

    def visit_Shift(self, binary):
        return self._a_pseudo_invocation()

    def visit_Additive(self, binary):
        return self._a_pseudo_invocation()

    def visit_Multiplicative(self, binary):
        return self._a_pseudo_invocation()

    def visit_MethodInvocation(self, invocation):
        return self._an_invocation()

    def visit_IfThenElse(self, conditional):
        return self._a_conditional()

    def visit_While(self, loop):
        return self._a_loop()

    def visit_For(self, loop):
        return self._a_loop()

    def visit_ForEach(self, loop):
        return self._a_loop()

    def visit_SwitchCase(self, conditional):
        return self._a_conditional()

    def visit_DoWhile(self, loop):
        return self._a_loop()

    def visit_Continue(self, statement):
        return self._a_pseudo_invocation()

    def visit_Break(self, statement):
        return self._a_pseudo_invocation()

    def visit_Throw(self, statement):
        return self._a_pseudo_invocation()

    def visit_ConstructorInvocation(self, invocation):
        return self._an_invocation()

    def visit_InstanceCreation(self, statement):
        return self._an_invocation()

    def visit_Literal(self, literal):
        return self._a_constant_value()

    # ----------

    def _end_of_method(self):
        self._uninitialised_locals = []
        return True

    def _can_have_final_locals(self):
        self._uninitialised_locals = []

    def _find_uninitialised(self, declaration, target):
        if 'final' in declaration.modifiers:
            for variable in declaration.variable_declarators:
                if not variable.initializer:
                    target.append(variable.variable.name)

    # ----------

    def _a_constant_value(self):
        self.components.constants += 1
        return True

    def _code_is_a_constant(self):
        self.components.constants += 1
        return True

    def _a_name(self):
        self.components.bindings += 1
        return True

    def _a_pseudo_invocation(self):
        # this is an invocation on object type Number, at least in Ruby
        self.components.invocations += 1
        return True

    def _an_invocation(self):
        self.components.invocations += 1
        return True

    def _a_conditional(self):
        self.components.conditionals += 1
        return True

    def _a_loop(self):
        self.components.loops += 1
        return True

    def _an_assignment(self):
        self.components.assignments += 1
        return True


def count_components(java_file):
    p = plyj.parser.Parser()
    tree = p.parse_file(java_file)
    visitor = CountingVisitor()
    tree.accept(visitor)
    return visitor.components
